#
# forecast.core - forecast basic infrastructure
#
# Copyright 2013-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import time
import urlparse

try:
	import json
except ImportError:
	import simplejson as json

from astro import Location

import asyn
import asyn.http

DEBUG = None


#
# System parameters.
#
APIHOST = 'api.darksky.net'


#
# A "data point" in forecast.io parlance
#
class Point(object):

	def __init__(self, data, units):
		if DEBUG: DEBUG("datapoint", data)
		self.units = units
		def convert(name, type, default=None):
			value = data.get(name, default)
			setattr(self, name, None if value is None else type(value))
		convert("time",					int)
		convert("summary",				unicode, "")
		convert("icon",					str, "unknown")
		convert("temperature",			float)
		convert("temperatureMin",		float)
		convert("temperatureMax",		float)
		convert("apparentTemperature",	float)
		convert("apparentTemperatureMin",float)
		convert("apparentTemperatureMax",float)
		convert("dewPoint",				float)
		convert("humidity",				float)
		convert("pressure",				float)
		convert("visibility",			float)
		convert("ozone",				float)
		convert("cloudCover",			float)
		convert("windSpeed",			float, 0)
		convert("windBearing",			float)
		convert("nearestStormDistance",	float)
		convert("nearestStormBearing",	float)
		convert("precipIntensity",		float, 0)
		convert("precipProbability",	float, 0)
		convert("precipType",			unicode, "")
		convert("precipAccumulation",	float)
		convert("temperatureMin",		float)
		convert("temperatureMinTime",	int)
		convert("temperatureMax",		float)
		convert("TemperatureMaxTime",	int)
		convert("moonPhase",			float)

	def __repr__(self):
		return "<Forecast Point@%s T%s H%s P%s V%s W%s@%s PR%s@%s/%s>" % (
			time.ctime(self.time),
			self.temperature, self.humidity, self.pressure, self.visibility,
			self.windSpeed, self.windBearing,
			self.precipIntensity, self.precipProbability, self.precipType
		)


#
# A vector of data points
#
class List(list):

	def __init__(self, data, units):
		if data is None:
			self.summary = None
			self.icon = None
			self.min = self.max = None
		else:
			self.summary = data.get("summary")
			self.icon = data.get("icon")
			points = data.get("data")
			if points:
				list.__init__(self, [Point(dp, units) for dp in points])
				times = [dp.time for dp in self]
				self.min = min(times)
				self.max = max(times)


#
# A weather alert
#
class Alert(object):

	def __init__(self, data):
		self.title = data.get("title")
		self.expires = data.get("expires")
		self.uri = data.get("uri")


#
# A full reading (reply) as delivered by darksky.net
#
class Reading(object):

	def __init__(self, data):
		self.raw = data
		s = json.loads(data)
		self.location = Location(s["latitude"], s["longitude"])
		self.alerts = [Alert(ad) for ad in s.get("alerts", [])]
		self.flags = s.get("flags")
		self.units = self.flags["units"]
		current = s.get("currently")
		self.current = Point(current, self.units) if current else None
		self.minutes = self._collect(s.get("minutely"))
		self.hours = self._collect(s.get("hourly"))
		self.days = self._collect(s.get("daily"))

	def getPoint(self, distance, units='days'):
		if units == 'now' or units is None:
			return self.current
		if units == 'days':
			base = self.days
		elif units == 'hours':
			base = self.hours
		else:
			return None
		if distance >= len(base):
			return None
		if DEBUG: DEBUG("get", distance, units, " => ", base[distance])
		return base[distance]
	
	def _collect(self, data):
		if data is not None:
			return List(data, self.units)
	
	def _alerts(self, data):
		if data is None:
			return []
		else:
			return [Alert(al) for al in data]

	def __repr__(self):
		return "<Reading%s current%s>" % (self.location, self.current)


#
# A communications channel to darksky.net.
#
class Forecast(asyn.Callable):
	""" Interface to the forecast.io service.

		This class contacts a fixed service vending forecast.io data and
		obtains one "reading", a present-time full data set at a given location.
	"""
	def __init__(self, control, callout=None):
		asyn.Callable.__init__(self, callout=callout)
		self.control = control
		self.apikey = None
		self.location = None
		self.user_agent = None
		self.units = 'us'

	def poll(self, callout):
		""" Explicitly get data from darksky.net, using preset parameters. """
		assert self.apikey is not None
		def cb(ctx, *args):
			if ctx.error:
				callout(ctx)
				req.close()
			elif ctx.state == 'body':
				if req.n_status == '200':
					callout(asyn.Context('reading'), Reading(args[0]))
				else:
					callout(asyn.Context('error'), req)
		query = dict(
			units=self.units
		)
		req = self._request("forecast", callout=cb, query=query)


	#
	# Web interface primitives
	#
	def _request(self, req, callout=None, action='GET', query=None):
		""" Send a web request to Dark Sky. """
		request = asyn.http.request(self.control, callout=callout, action=action, query=query, compression=False)
		if self.user_agent:
			request.user_agent = self.user_agent + ' ' + request.user_agent
		request.open(self._weburl(req))
		return request

	def _weburl(self, op):
		return urlparse.urlunsplit(('https', APIHOST,
			"/%s/%s/%s,%s" % (op, self.apikey, self.location.lat, self.location.lon),
			None, None))


#
# Regression test
#
if __name__ == "__main__":
	import getopt, sys

	def cb(ctx, arg = None):
		print "MAIN CALLOUT", ctx, arg

	def dlog(*it):
		print ' '.join(map(str, it))

	opts, args = getopt.getopt(sys.argv[1:], "H")
	for opt, value in opts:
		if opt == '-H':
			asyn.http.DEBUG = dlog
	control = asyn.Controller()
	forecast = Forecast(control, callout=cb)
	forecast.apikey = args[0]
	forecast.location = Location(85, 90)
	def rep(ctx, arg=None):
		if ctx.error:
			print "ERROR", ctx.error
		elif ctx.state == 'error':
			print "HTTP", arg.n_status, arg.v_status
		elif ctx.state == 'reading':
			print "Reading:", arg
		else:
			print "UNEXPECTED", ctx, arg
		control.close()
	forecast.poll(rep)
	control.run()
